// The following code runs speed benchmarks for the 1:N identification module

#include <chrono>
#include <iostream>
#include <string>
#include <vector>

#include "tf_sdk.h"

typedef std::chrono::high_resolution_clock Clock;


int main() {
    // TODO: Replace with your license
    const std::string license = "${LICENSE_CODE}";

    Trueface::ConfigurationOptions options;
    options.frModel = Trueface::FacialRecognitionModel::FULL;
    options.dbms = Trueface::DatabaseManagementSystem::NONE;

    // TODO modify the following option to test with / without vector compression
    options.frVectorCompression = true;

    if (options.frVectorCompression) {
        std::cout << "~~~~~~~~~~~~~~~ Vector compression enabled ~~~~~~~~~~~~~~~~"  << std::endl;
    }

    if (options.frModel == Trueface::FacialRecognitionModel::FULL) {
        std::cout << "~~~~~~~~~~~~~~~ FULL model ~~~~~~~~~~~~~~~~"  << std::endl;
    } else {
        std::cout << "~~~~~~~~~~~~~~~ LITE model ~~~~~~~~~~~~~~~~"  << std::endl;
    }

    Trueface::SDK sdk(options);
    bool valid = sdk.setLicense(license);
    if (!valid) {
        std::cout << "License not valid" << std::endl;
        return -1;
    }

    // Populate our collection with the following distractor templates
    std::vector<std::pair<std::string, std::string>> imagePairList = {
            {"../../images/armstrong_2.jpg", "armstrong"},
            {"../images/headshot.jpg", "anon"}
    };

    std::vector<std::pair<std::string, Trueface::Faceprint>> dataVec;

    // Generate a template for each of the images
    std::cout << "\nGenerating templates" << std::endl;
    for (const auto& imagePair: imagePairList) {
        auto ret = sdk.setImage(imagePair.first);
        if (ret != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to set image: " << imagePair.first << std::endl;
            return -1;
        }

        Trueface::Faceprint faceprint;
        ret = sdk.getLargestFaceFeatureVector(faceprint);

        if (ret != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to extract features" << std::endl;
            return -1;
        }

        dataVec.emplace_back(imagePair.second, std::move(faceprint));
    }

    // Prepare the probe vector and the match vector
    auto ret = sdk.setImage("../../images/aldrin.jpg");
    if (ret != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to set image" << std::endl;
        return -1;
    }

    Trueface::Faceprint probe;
    ret = sdk.getLargestFaceFeatureVector(probe);
    if (ret != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to extract features" << std::endl;
        return -1;
    }

    ret = sdk.setImage("../../images/aldrin_2.jpg");
    if (ret != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to set image" << std::endl;
        return -1;
    }

    Trueface::Faceprint matchTemplate;
    ret = sdk.getLargestFaceFeatureVector(matchTemplate);
    if (ret != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to extract features" << std::endl;
        return -1;
    }

    // Collection sizes to test
    std::vector<size_t> collectionSizes {
        1000,
        10000,
        100000,
        1000000
    };

    // Populate the collections
    for (const auto& collectionSize: collectionSizes) {
        std::cout << "Populating collection with " << collectionSize << " templates" << std::endl;

        // We are using the Trueface::DatabaseManagementSystem::NONE so the collection will not persist
        ret = sdk.createLoadCollection("temp_collection");
        if (ret != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Error creating collection" << std::endl;
            return -1;
        }

        // Enroll the templates
        // Note: we are enrolling collectionSize - 1 because we will add the matchTemplate at the end.
        for (size_t i = 0; i < collectionSize; ++i) {

            const auto &data = dataVec[i % dataVec.size()];
            std::string UUID;

            // Enroll the template and the identity
            ret = sdk.enrollTemplate(data.second, data.first, UUID);
            if (ret != Trueface::ErrorCode::NO_ERROR) {
                std::cout << "Unable to enroll template" << std::endl;
                return -1;
            }
        }

        // Finally enroll the match template
        std::string UUID;
        ret = sdk.enrollTemplate(matchTemplate, "aldrin", UUID);
        if (ret != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to enroll template" << std::endl;
            return -1;
        }

        // Run the timing tests
        size_t numIterations = 1000;
        if (collectionSize >= 100000) {
            numIterations = 100;
        }

        Trueface::Candidate candidate;
        bool found;

        auto t1 = Clock::now();
        for (size_t i = 0; i < numIterations; ++i) {
            sdk.identifyTopCandidate(probe, candidate, found);
        }
        auto t2 = Clock::now();

        if (found) {
            std::cout << "Found match: " << candidate.identity << std::endl;
        }

        double avgTime = static_cast<double>(std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count()) / numIterations;
        std::cout << "[Identify] Collection size: " << collectionSize << " templates, Avg time: " << avgTime << " milliseconds ("
                  << std::to_string(numIterations) << " iterations)\n" << std::endl;

    }

    return 0;
}

















